﻿--local StatLogic = LibStub:GetLibrary("LibStatLogic-1.1")
local StatLogic = LibStub("LibStatLogic-1.2")
local Toolkit = LibStub("Toolkit-1.0", 2);

HealPointsCalculator ={
	HP_LastPowerfulCurrent = nil;
	--HP_LastPowerfulNew = nil;
	HP_LastEfficientCurrent = nil;
--	HP_LastEfficientNew = nil;
};

--[[
	function GetHPTick	
--]]

local BonusPercentInt;
local BonusPercentSpi;
local _cachedCurrentPlayerStats = nil; --HealPointsCalculator:GetCurrentPlayerStats() return this cached value
local _cachedCurrentHealPoints = nil; --HealPointsCalculator:GetCurrentHealPoints(nil) returns this cached value


-- Helper functions
--[[
	Returns the global cooldown time (normally 1.5 seconds) after
	it has been reduced by specifid amount of spell haste
--]]
local function getGCD(spellHaste)
	if (spellHaste == nil) then
		return 1.5;
	else
		return math.max(1.0, 1.5 / (1 + spellHaste / 100));
	end
end

--[[
	Get the amount of healing done by a particular spell
	@param spell A HealPointsSpells.SPELLTABLE entry the indicates the spell to be calculated
	@param healing A bonus amount of healing to be added to the spell
	@param spellCrit The healing crit chance as a percentage  e.g. 11.7 ==> 11.7% = 0.117
--]]
local function getHPCast(spell, healing, spellCrit)
	assert(spell, "calculator.lua:getHPCast - spell is nil");
	assert(healing, "calculator.lua:getHPCast - healing is nil");
	assert(spellCrit, "calculator.lua:getHPCast - spellCrit is nil");

	HealPoints:LevelDebug(2, "getHPCast: spell=%s", Toolkit:VarAsString(spell));


	local spellType = spell.class.type;


		
	local percentBonus = HealPointsSpells:getHealingPercentBonus(spell);
	local healFactor, healFactor2 = HealPointsSpells:getHealingFactor(spell);
	spellCrit = spellCrit + HealPointsSpells:getSpellCritBonus(spell);
	healing = healing + HealPointsSpells:getHealingBonus(spell);

	local out, minHeal, maxHeal;


	if ((spellType == 'normal') or (spellType == 'instant')) then
		--HealPoints:LevelDebug(2, "   getHPCast (Normal,Instant): percentBonus=%s", Toolkit:VarAsString(percentBonus));
		--HealPoints:LevelDebug(2, "   getHPCast (Normal,Instant): healFactor=%s", Toolkit:VarAsString(healFactor));
		--HealPoints:LevelDebug(2, "   getHPCast (Normal,Instant): spellCrit=%s (after spell crit bonus)", Toolkit:VarAsString(spellCrit));
		--HealPoints:LevelDebug(2, "   getHPCast (Normal,Instant): healing=%s (after spell healing bonus)", Toolkit:VarAsString(healing));

		--minHeal = (spell['min'] + healing * healFactor) * percentBonus * (1 + spellCrit / 200);
		--maxHeal = (spell['max'] + healing * healFactor) * percentBonus * (1 + spellCrit / 200);

		--20120723: It's base*percentBonus + bonusHealingFromSpellPower     then   * spellCrit
		--20120723: Spell criticals cause double heal (e.g. 6192 -> 12573)

		minHeal = ( spell['min']*percentBonus + healing*healFactor) * (1 + spellCrit / 100);
		maxHeal = ( spell['max']*percentBonus + healing*healFactor) * (1 + spellCrit / 100);

		out = (minHeal+maxHeal) / 2.0;
	elseif (spell.class['type'] == 'hot') then
		local healTick = (spell['tick']*percentBonus + healing*healFactor);

		out = healTick*HealPointsSpells:getNumberOfTicks(spell);

		minHeal = out;
		maxHeal = out;
	elseif (spell.class['type'] == 'regrowth') then

		-- Burst part
		minHeal = (spell['min']*percentBonus + healing*healFactor) * (1 + spellCrit / 100);
		maxHeal = (spell['max']*percentBonus + healing*healFactor) * (1 + spellCrit / 100);

		-- HOT part
		local healTick = (spell['tick']*percentBonus + healing*healFactor2);

		minHeal = minHeal + healTick*HealPointsSpells:getNumberOfTicks(spell);
		maxHeal = maxHeal + healTick*HealPointsSpells:getNumberOfTicks(spell);

		out = (minHeal+maxHeal) / 2;
	elseif (spell.class['type'] == 'lifebloom') then
		--todo: does +healing affect lifebloom style burst or hot parts?

		-- HOT part
		local healTick = spell['tick']*percentBonus;
		out = healTick*HealPointsSpells:getNumberOfTicks(spell);

		-- Burst part
		local burseHeal = spell['burst']*percentBonus * (1 + spellCrit / 200);
		
		out = out + burstHeal;
	elseif (spell.class['type'] == 'chain') then
		minHeal = (spell['min'] * percentBonus + healing*healFactor) * (1 + spellCrit / 100);
		maxHeal = (spell['max'] * percentBonus + healing*healFactor) * (1 + spellCrit / 100);
		--out = (((spell['min'] + spell['max']) / 2) + healing * healFactor) * percentBonus * (1 + spellCrit / 200);

		--local multiTargetFactor = 2.2; --(1+0.7+0.4+0.1);
		--local multiTargetFactor = 2.533; --assume four targets (1+0.7+0.4+0.1)
		--local multiTargetFactor = 2.1; --assume three targets (1+0.7+0.4)
		local multiTargetFactor = 1.7; --assume two targets only (1+0.7)

		minHeal = minHeal * multiTargetFactor; -- Assumes that the spell hits all 3 targets
		maxHeal = maxHeal * multiTargetFactor; -- Assumes that the spell hits all 3 targets

		out = (minHeal + maxHeal) / 2;

	elseif (spell.class['type'] == 'aoe') then
--		out = (((spell['min'] + spell['max']) / 2) + healing * healFactor) * percentBonus * (1 + spellCrit / 200)

		minHeal = (spell['min'] * percentBonus + healing*healFactor) * (1 + spellCrit / 100);
		maxHeal = (spell['max'] * percentBonus + healing*healFactor) * (1 + spellCrit / 100);

		assert(minHeal ~= 0, string.format("getHPCast: aoe (%s) minHeal is zero", spell.class.abbr));
		assert(maxHeal ~= 0, string.format("getHPast: aoe (%s) maxHeal is zero", spell.class.abbr));

		local nTargets = math.min(5, HealPoints.db.char.hot['numtargets']);
		if (nTargets < 1) then
			print(string.format("HealPoints.db.char.hot['numtargets'] (%d) is less than 1. Assuming 1 target", nTargets));
			nTargets = 1;
		end;
		minHeal = minHeal * nTargets; -- Assumes that the spell hits all n targets
		maxHeal = maxHeal * nTargets; -- Assumes that the spell hits all n targets

		out = (minHeal + maxHeal) / 2;
	end

	--If Divine Plea is used it decreases our healing by 50% while it's active
	if (_G["HealPointsOptionsConfigDivinePlea"]:GetChecked() == 1) then -- Divine plea
		-- 50% reduction 9/120 of the time = 3.75% reduction on average
		out = out * 0.9625;
	end

	HealPoints:LevelDebug(2, "getHPCast: result=%s, %s - %s", 
			Toolkit:VarAsString(out),
			Toolkit:VarAsString(minHeal),
			Toolkit:VarAsString(maxHeal));


  	return out, minHeal, maxHeal;
end

--[[
	Get HP healed per tick of a HOT spell
--]]
function HealPointsCalculator:GetHPTick(spell, healing)
	local percentBonus = HealPointsSpells:getHealingPercentBonus(spell);
	local healFactor, healFactor2 = HealPointsSpells:getHealingFactor(spell);
	healing = healing + HealPointsSpells:getHealingBonus(spell);

	local healTick = 0;

	if (spell.class['type'] == 'hot') then
		healTick = spell['tick'] + healing * healFactor / spell.class['tickcount']; -- +healing divided between original tickcount
	elseif (spell.class['type'] == 'regrowth') then
		healTick = spell['tick'] + healing * healFactor2 / spell.class['tickcount'];
	elseif (spell.class['type'] == 'lifebloom') then
		healTick = 
				spell['tick']
				--+ (healing + 2.777 * HealPointsBS:GetBonus('AVG_HOT_LIFEBL')) * healFactor / spell.class['tickcount'];
	elseif (spell.class['type'] == 'aoe') then
		healTick = spell['tick'] + healing*healFactor;
	end

	healTick = healTick * percentBonus;
	return healTick;
end

--[[
	Get the time between casts.
	Depending on the spell you might be limited by:
		- GCD
		- cast time
		- cooldown
	
	Parameters
		spell (table, SPELLTABLE entry): A HealPointsSpells.SPELLTABLE entry that is the spell we want to get the cast times of
		spellHaste (number, percentage): A percentage spell haste that speeds up the spell's cast time

	Returns
		castTime (number, seconds): The minimum amount of time in seconds the player must wait between casting the spell (GCD if the spell is instant)
		recastTime (number, seconds): The amount of time in seconds the player would wait to recast the spell on the same person/group (if it's an over-time spell)		
--]]
local function getSpellCastTimes(spell, spellHaste)
	assert(spell);
	assert(spellHaste);
	--assert(extendToIncludeOverTimeComponent);
--	self:argCheck(spell, 1, "table");
--	self:argCheck(spellHaste, 2, "number");
	
	local castTime = HealPointsSpells:GetCastTime(spell, spellHaste);
	assert(castTime ~= nil, string.format("HealPointsSpells:GetCastTime(spell, %d) returned nil", spellHaste));
	
	if (castTime == 0) then
		castTime = getGCD(spellHaste);
	end;
	
	if (spell.class['cooldown'] ~= nil) then
		castTime = math.max(castTime, spell.class['cooldown']);
	end;

	local castTimeWithHotTime;
	
	if ((spell.class['interval'] or 0) > 0) then
		local hotTime = spell.class['interval'] * HealPointsSpells:getNumberOfTicks(spell);
		castTimeWithHotTime = castTime + hotTime;
	end;

	return castTime, castTimeWithHotTime;
end;

--[[
	Calculate the amount of healing per second done by a spell
--]]
local function getHPSec(spell, healing, spellCrit, spellHaste)

	local hpPerCast = getHPCast(spell, healing, spellCrit);
	local castTime, castTimeWithHot = getSpellCastTimes(spell, spellHaste);

	--[[
	if (spell.class['type'] == 'normal' or spell.class['type'] == 'chain' or spell.class['type'] == 'aoe') then
		castTime = HealPointsSpells:GetCastTime(spell, spellHaste);
		if (castTime == 0) then
			castTime = getGCD(spellHaste);
		end
		if (spell.class['cooldown'] ~= nil) then
			castTime = math.max(castTime, spell.class['cooldown']);
		end
	elseif (spell.class['type'] == 'instant') then
		castTime = spell.class['cooldown'];
	elseif (spell.class['type'] == 'hot' or spell.class['type'] == 'lifebloom') then
		castTime = (HealPointsSpells:getNumberOfTicks(spell) * spell.class['interval']);
	elseif (spell.class['type'] == 'regrowth') then
		castTime = (HealPointsSpells:GetCastTime(spell, spellHaste) + HealPointsSpells:getNumberOfTicks(spell) * spell.class['interval']);
	end
	--]]

	return hpPerCast/castTime;
end

--[[
	Calcuate the amount of healing done per point of mana by a spell
--]]
local function getHPMana(spell, healing, spellCrit)
	--HealPoints:LevelDebug(2, "getHPMana: spell=%s", Toolkit:VarAsString(spell));
	--HealPoints:LevelDebug(2, "getHPMana: healing=%s", Toolkit:VarAsString(healing));
	--HealPoints:LevelDebug(2, "getHPMana: spellCrit=%s", Toolkit:VarAsString(spellCrit));

	local hpcast = getHPCast(spell, healing, spellCrit);
	--HealPoints:LevelDebug(2, "   getHPMana: hpcast=%s", Toolkit:VarAsString(hpcast));

	local manaCost = HealPointsSpells:getManaCost(spell, spellCrit);
	HealPoints:LevelDebug(2, "   getHPMana: manaCost=%s", Toolkit:VarAsString(manaCost));

	return hpcast / manaCost;
end

local function getHPTotalRegen(spell, healing, spellCrit, spellHaste, combatRegen, mana)
	local hpPerCast = getHPCast(spell, healing, spellCrit);
	local manaCost = HealPointsSpells:getManaCost(spell, spellCrit);

	local interval; --how often the spell can be cast
	if (spell.class['type'] == 'regrowth') then
		interval = HealPointsSpells:GetCastTime(spell, spellHaste) + HealPointsSpells:getNumberOfTicks(spell) * spell.class['interval'];
	elseif (spell.class['type'] == 'hot' or spell.class['type'] == 'lifebloom') then
		interval = HealPointsSpells:getNumberOfTicks(spell) * spell.class['interval'];
	elseif (spell.class['type'] == 'instant') then
		interval = math.max(getGCD(spellHaste), spell.class['cooldown']);
	else
		interval = math.max(getGCD(spellHaste), HealPointsSpells:GetCastTime(spell, spellHaste)); 
	end

	if hpPerCast == 0 then
		error(string.format("HealPoints_Calculator getHPTotalRegen: hpPerCast is zero for spell %s", spell.class['type']));
	end;

	local manaUsageSpell = manaCost - interval * combatRegen;

	local hpPerMana;

	--gotta catch division by zero.
	--LUA doesn't define if 6000/0 is +INF or NaN
	--Internally we use negative values to mean infinity
	if (manaUsageSpell > 0) then
		hpPerMana = hpPerCast / manaUsageSpell;
	else
		hpPerMana = -1;
	end;

	local hpPerManaBar = hpPerMana * mana;

	if isnan(hpPerManaBar) then
		error(string.format("HealPoints_Calculator hpPerManaBar is NaN for spell %s. hpPerCast=%s, manaUsageSpell=%s", 
					spell.class['type'], hpPerCast, manaUsageSpell));	
	end;
	
	local numberCasts = mana / manaUsageSpell;
	local totalTime = numberCasts * interval;

	return hpPerManaBar, totalTime;
end

--[[
	Calculate the total amount of healing that can be done by a spell in the specified amount of time, starting at the specified mana level
	@param spell - A SPELLTABLE entry of the spell to calculate
	@param healing - A number that is an amount of bonus healing
	@param seconds - How long we run the calculation for
	@param startMana - The player's starting mana
	@param spellCrit - The player's chance of getting a critical heal
	@param spellHaste - The player's spell haste
	@param combatRegen - The players mana regeneration per second while in combat
	--@param manaRegenNormal - The players mana regeneration per second while out of combat
--]]
local function computeTotalHealingNormal(spell, healing, seconds, startMana, spellCrit, spellHaste, combatRegen)
	--DEFAULT_CHAT_FRAME:AddMessage("Computing total healing for "..spell.class['name'].."("..spell['rank']..") secs="..seconds.." mana="..startMana,1,1,1);

	assert(spell);
	assert(healing);
	assert(seconds);
	assert(startMana);
	assert(spellCrit);
	assert(spellHaste);
	assert(combatRegen);
	--assert(manaRegenNormal);

	--simple bailout. If they configured the time for zero seconds then we couldn't heal anything
	if (seconds == 0) then
		return 0;
	end;

	--record the details of our calculations
	local details = {};
		--hpCast Health Points healed by a cast of the spell
		--castDelay Time the player must wait between casts (taking into effect cast time, cooldown, haste and gcd)
		--manaCost The effective mana cost of the spell (some players can have talents to refund mana, reducing effective mana cost)
		--hps Healing per second. Literally hpCast/castDelay. This number only makes sense while the player has mana to power through chain-cast
		--manaUsageSecond Mana used per second while chain-casting (takes into account that you can regen while casting)
		--secondsPhase1 Number of seconds the player can chain-cast before they run out of mana
		--totalHealedPhase1 The total amount of health healed in phase 1 chain-casting
		--secondsPhase2 Number of seconds that the player has to limp along waiting for mana to regen each cast, before time is up
		--cycleTime Number of seconds the player will have to wait between casts, as they wait for mana to regen
		--spellsCast
		--totalHealedPhase2 The total amount of health healed in phase 2 OOM casting


	local hpCast = getHPCast(spell, healing, spellCrit); --amount of HP healed per cast
	local castDelay = math.max(getGCD(spellHaste), HealPointsSpells:GetCastTime(spell, spellHaste)); --how fast i can cast it
	if (spell.class['cooldown'] ~= nil) then
		castDelay = math.max(castDelay, spell.class['cooldown']);
	end

	local manaCost = HealPointsSpells:getManaCost(spell, spellCrit);
	local hps = hpCast / castDelay;

	HealPoints:LevelDebug(1, "      computeTotalHealingNormal: spell=%s, hpCast=%d, manaCost=%d, castDelay=%s, combatRegen=%d",
			spell.class.abbr, hpCast, manaCost, castDelay, combatRegen);

	-- Phase 1: Has mana to chain-cast
	local manaUsageSecond = max( (manaCost / castDelay) - combatRegen, 0); --in case mana regen exceeds mana cost

	HealPoints:LevelDebug(1, "      computeTotalHealingNoraml: manaUsageSecond=%d", manaUsageSecond);

	--HealPoints:LevelDebug(3, "combatRegen=%s, manaUsageSecond=%s", combatRegen, manaUsageSecond);

	HealPoints:LevelDebug(3, "[%s] (%.2f seconds) startMana=%.2f, manaUsageSecond=%.2f, manaCost=%.2f, castDelay=%.2f, combatRegen=%.2f",
			spell.class.abbr, seconds, startMana, manaUsageSecond, manaCost, castDelay, combatRegen);

	--If mana usage per second is zero, then phase 1 lasts as long as the configured time
	local secondsPhase1;
	if (manaUsageSecond > 0) then
		secondsPhase1 = math.min(seconds, startMana / manaUsageSecond);
	else 
		secondsPhase1 = seconds;
	end;
	local totalHealed = secondsPhase1 * hps;

	HealPoints:LevelDebug(3, "[%s] Phase 1 (%s s): totalHealed=%.2f, hps=%.2f", 
			spell.class.abbr, secondsPhase1, totalHealed, hps);



	-- Phase 2: Has no mana left
	local secondsPhase2 = seconds - secondsPhase1;
	if (secondsPhase2 > 0) then
		-- Seconds between each casting
		local cycleTime = math.max(castDelay, manaCost / combatRegen);
		local spellsCast = secondsPhase2 / cycleTime;
		local healedPhase2 = spellsCast * hpCast;
		totalHealed = totalHealed + healedPhase2;

		HealPoints:LevelDebug(3, "Phase 2: %s: totalHealed2=%.2f, secondsPhase2=%.2f, cycleTime=%.2f",
				spell.class.abbr, healedPhase2, secondsPhase2, cycleTime);
	end

	return totalHealed;
end

local function computeTotalHealingRegrowth(spell, healing, seconds, startMana, spellCrit, spellHaste, combatManaRegen)
	assert(spell);
	assert(healing);
	assert(seconds);
	assert(startMana);
	assert(spellCrit);
	assert(spellHaste);
	assert(combatManaRegen);
	--assert(manaRegenNormal);

	HealPoints:LevelDebug(3, "         computeTotalHealingRegrowth(spell=%s, healing=%d, seconds=%d, startMana=%d, spellCrit=%d, spellHaste=%d, combatManaRegen=%d", 
			Toolkit:VarAsString(spell), healing, seconds, startMana, spellCrit, spellHaste, combatManaRegen);

	--DEFAULT_CHAT_FRAME:AddMessage("Computing total healing for "..spell.class['name'].."("..spell['rank']..") secs="..seconds.." mana="..startMana,1,1,1);
	local manaCost = HealPointsSpells:getManaCost(spell, spellCrit);
	--clit

	--local castTime = math.max(getGCD(spellHaste), HealPointsSpells:GetCastTime(spell, spellHaste));
	local castTime, castTimeWithHot = getSpellCastTimes(spell, spellHaste); 
	local percentBonus = HealPointsSpells:getHealingPercentBonus(spell);
	local healFactorBurst, healFactorHot = HealPointsSpells:getHealingFactor(spell);
	local numtargets = HealPoints.db.char.hot['numtargets'];
	
	HealPoints:LevelDebug(2, "            computeTotalHealingRegrowth: numtargets=%d", numtargets);
	HealPoints:LevelDebug(2, "            computeTotalHealingRegrowth: manaCost=%d", manaCost);
	HealPoints:LevelDebug(2, "            computeTotalHealingRegrowth: castTime=%d", castTime);

	spellCrit = spellCrit + HealPointsSpells:getSpellCritBonus(spell);
	healing = healing + HealPointsSpells:getHealingBonus(spell);

	--clit2
	local avgBurst = (((spell['min'] + spell['max']) / 2) + healing * healFactorBurst) * percentBonus * (1 + spellCrit / 200);
	local healTick = (spell['tick'] + healing * healFactorHot / spell.class['tickcount']) * percentBonus;
	

	-- Phase 1: Has mana to chain-cast
	local manaUsagePerSecond = (manaCost / castTime) - combatManaRegen;
	local secondsPhase1 = math.min(seconds, startMana / manaUsagePerSecond);

	HealPoints:LevelDebug(2, "            computeTotalHealingRegrowth: startMana=%d, manaCost=%d, manaUsagePerSecond=%d, secondsPhase1=%d", 
			startMana, manaCost, manaUsagePerSecond, secondsPhase1);

	local secsBetweenSameTarget = math.max(castTime, numtargets * castTime);
	local ticsPerCast = secsBetweenSameTarget / spell.class['interval'];
	ticsPerCast = math.min(ticsPerCast, HealPointsSpells:getNumberOfTicks(spell));
	local spellsCast = secondsPhase1 / castTime;

	local totalHealed = spellsCast * (avgBurst + ticsPerCast * healTick);

	-- Phase 2: Has no mana left
	local secondsPhase2 = seconds - secondsPhase1;
	if (secondsPhase2 > 0) then
		local cycleTime =  math.max(castTime, manaCost / combatManaRegen); -- Seconds between each possible cast on different targets
		--local playerTime = math.max(


		secsBetweenSameTarget = math.max(cycleTime, numtargets * cycleTime);
		ticsPerCast = secsBetweenSameTarget / spell.class['interval'];
		ticsPerCast = math.min(ticsPerCast, HealPointsSpells:getNumberOfTicks(spell));
		spellsCast = secondsPhase2 / cycleTime;

		local healedPhase2 = spellsCast * (avgBurst + ticsPerCast * healTick);
		totalHealed = totalHealed + healedPhase2;
	end

	return totalHealed;
end

local function computeTotalHealingLifebloom(spell, healing, seconds, startMana, spellCrit, spellHaste, combatRegen)

	--Get the the percent bonus to healing spells (e.g. 1.07 = 7%)
	local percentBonus = HealPointsSpells:getHealingPercentBonus(spell);
	local healFactorHot, healFactorBurst = HealPointsSpells:getHealingFactor(spell);
	spellCrit = spellCrit + HealPointsSpells:getSpellCritBonus(spell);
	healing = healing + HealPointsSpells:getHealingBonus(spell);
	
	local avgBurst = (
			spell['burst'] 
			--+ (healing + HealPointsBS:GetBonus('AVG_BURST_LIFEBL')) * healFactorBurst
		) * percentBonus;
		
	local healTick = (
		spell['tick'] 
		--+ (healing + HealPointsBS:GetBonus('AVG_HOT_LIFEBL')) * healFactorHot / spell.class['tickcount']
		) * percentBonus;
	avgBurst = avgBurst * (1 + spellCrit / 200);
	local duration = HealPointsSpells:getNumberOfTicks(spell) * spell.class['interval'];
	local manaCost = HealPointsSpells:getManaCost(spell, spellCrit);
	local numtargets = HealPoints.db.char.hot['numtargets'];

	if (_G["HealPointsOptionsConfigLifeBl"]:GetChecked() == 1) then -- stack lifeblooms
		-- Phase 1: Has enough mana/regen to keep stacking lifeblooms up on numtargets constantly
		numtargets = math.min(3, numtargets); -- max 3 targets
		local cycleTime = duration;
		local manaUsageCycle = numtargets * manaCost;
		local manaRegainCycle = cycleTime * combatRegen;

		local numCycles = seconds / cycleTime;
		if (manaRegainCycle < manaUsageCycle) then -- if you regen more than you use, you'll never go oom.
			numCycles = math.min(numCycles, startMana / (manaUsageCycle - manaRegainCycle));
		end
		local secondsPhase1 = numCycles * cycleTime;
		local numTicks = numCycles * numtargets * 3 * (HealPointsSpells:getNumberOfTicks(spell) - 1); -- Overwrites before the last tick
		if (numCycles > 2) then -- Ticks lost due to building stacks
			numTicks = numTicks - numtargets * 3 * (HealPointsSpells:getNumberOfTicks(spell) - 1) - 3 * 0.75 * (numtargets * numtargets - numtargets);
		elseif (numCycles == 2) then
			numTicks = numTicks - numtargets * 3 * (HealPointsSpells:getNumberOfTicks(spell) - 1) - 2 * 0.75 * (numtargets * numtargets - numtargets);
		elseif (numCycles == 1) then
			numTicks = numTicks - numtargets * 2 * (HealPointsSpells:getNumberOfTicks(spell) - 1) - 0.75 * (numtargets * numtargets - numtargets);
		end
		local totalHealed = numTicks * healTick;
		if (numCycles > 1) then
			totalHealed = totalHealed + numtargets * (avgBurst + healTick); -- Last tick + burst when oom
		end

		-- Phase 2: Has no mana left, no stacking
		local secondsPhase2 = seconds - secondsPhase1;
		if (secondsPhase2 > 0) then
			cycleTime = manaUsageCycle / combatRegen; --how long it takes to get enough mana to cast a cycle
			numCycles = secondsPhase2 / cycleTime;
			local healedPhase2 = numCycles * numtargets * (avgBurst + healTick * HealPointsSpells:getNumberOfTicks(spell));
			totalHealed = totalHealed + healedPhase2;
		end
		return totalHealed;
	
	else
		-- Phase 1: Has enough mana/regen to keep lifebloom up on numtargets constantly
		local cycleTime = math.max(duration + 1, numtargets * getGCD(spellHaste)); -- time between casts on the same target
		local manaUsageCycle = numtargets * manaCost;
		local regenTimeCasting = math.min(cycleTime, 5 + (numtargets - 1) * getGCD(spellHaste));
		local regenTimeNormal = math.max(0, cycleTime - regenTimeCasting);
		local manaRegainCycle = regenTimeCasting * (combatRegen / 5) + regenTimeNormal * (manaRegenNormal / 5);
		local healingPerCycle = numtargets * (avgBurst + healTick * HealPointsSpells:getNumberOfTicks(spell));
		local numCycles = seconds / cycleTime;
		if (manaRegainCycle < manaUsageCycle) then -- if you regen more than you use, you'll never go oom.
			numCycles = math.min(numCycles, startMana / (manaUsageCycle - manaRegainCycle));
		end
		local secondsPhase1 = numCycles * cycleTime;
		local totalHealed = numCycles * healingPerCycle;

		-- Phase 2: Has no mana left
		local secondsPhase2 = seconds - secondsPhase1;
		if (secondsPhase2 > 0) then
			local manaToRegen = manaUsageCycle - regenTimeCasting * (combatRegen / 5);
			cycleTime = regenTimeCasting + manaToRegen / (manaRegenNormal / 5);
			numCycles = secondsPhase2 / cycleTime;
			local healedPhase2 = numCycles * healingPerCycle;
			totalHealed = totalHealed + healedPhase2;
		end

		return totalHealed;
	end
end

local function computeTotalHealingHot(spell, healing, seconds, startMana, spelllHaste, combatRegen)
	local hpCast = getHPCast(spell, healing, 0); -- Can't crit
	local manaCost = HealPointsSpells:getManaCost(spell, 0); -- Can't crit
	local duration = HealPointsSpells:getNumberOfTicks(spell) * spell.class['interval'];
	local numtargets = HealPoints.db.char.hot['numtargets'];
	local cycleTime = math.max(duration + 1, numtargets * getGCD(spellHaste)); -- time between casts on the same target
	local healingPerCycle = numtargets * hpCast; --total amount healed per cycle
	local manaUsageCycle = numtargets * manaCost; --mana lost per cycle
	local manaRegainCycle = cycleTime * combatRegen; --mana regenerated per cycle

	--local regenTimeCasting = math.min(cycleTime, 5 + (numtargets - 1) * getGCD(spellHaste));
	--local regenTimeNormal = math.max(0, cycleTime - regenTimeCasting);

	-- Phase 1: Has enough mana/regen to keep the HoT up on numtargets constantly
	local numCycles = seconds / cycleTime;
	-- if you regen more than you use, you'll never go oom.
	if (manaRegainCycle < manaUsageCycle) then 
		numCycles = math.min(numCycles, startMana / (manaUsageCycle - manaRegainCycle));
	end
	local secondsPhase1 = numCycles * cycleTime;
	local totalHealed = numCycles * healingPerCycle;

	-- Phase 2: Has no mana left
	local secondsPhase2 = seconds - secondsPhase1;
	if (secondsPhase2 > 0) then
		cycleTime = max(cycleTime, manaUsageCycle/combatRegen);
		numCycles = secondsPhase2 / cycleTime;
		local healedPhase2 = numCycles * healingPerCycle;
		totalHealed = totalHealed + healedPhase2;
	end

	return totalHealed;
end

--[[
	Caluclates one of the two terms in the HealPoints equation (HealPoints = PowerPoints+EndurancePoints
	Calculates how much healing can be done by an entire mana bar; takes mana regen into account
	@param maxMana: number - The player's current percent [0..100] of their mana
--]]
local function computePowerPoints(healing, spellCrit, spellHaste, maxMana, combatRegen)

	assert(healing, "computePowerPoints - healing is nil");
	assert(spellCrit, "computePowerPoints - spellCrit is nil");
	assert(spellHaste, "computePowerPoints - spellHaste is nil");
	assert(maxMana, "computePowerPoints - maxMana is nil");
	assert(combatRegen, "computePowerPoints - combatRegen is nil");
	--assert(manaRegenNormal, "computePowerPoints - manaRegenNormal is nil");

	local mana = HealPoints.db.char.power['mana'] * maxMana / 100;

	HealPoints:LevelDebug(1, "   computePowerPoints(healing=%d, spellCrit=%d, spellHaste=%d, maxMana=%d, combatRegen=%d, mana=%d", 
				healing, spellCrit, spellHaste, maxMana, combatRegen, mana);

	if (HealPoints.db.char.power['auto'] == true) then
		local powerfulSpell;
		local maxPowerPoints = 0;
		for i = 1, table.getn(HealPointsSpells.SPELLTABLE) do
			for j = 1, table.getn(HealPointsSpells.SPELLTABLE[i]), 1 do
				local spell = HealPointsSpells.SPELLTABLE[i][j];

				local spellType = spell.class['type'];

				HealPoints:LevelDebug(1, "      computerPowerPoints: %s (%s)", Toolkit:VarAsString(spell.class.abbr), spellType);

				if (spell ~= nil and HealPointsSpells:getHighestSpellRank(spell.class.classID) == spell['rank'] and spell.class['type'] ~= 'instant') then
					local powerPoints;
					if (spellType == 'hot') then
						powerPoints = computeTotalHealingHot(spell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellHaste, combatRegen);
					elseif (spellType == 'regrowth') then
						powerPoints = computeTotalHealingRegrowth(spell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, spellHaste, combatRegen);
					elseif (spellType == 'lifebloom') then
						powerPoints = computeTotalHealingLifebloom(spell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, spellHaste, combatRegen);
					else
						powerPoints = computeTotalHealingNormal(spell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, spellHaste, combatRegen);
					end
					
					HealPoints:LevelDebug(1, "         computePowerPoints: %s=%d (%s)", spell.class['abbr'], powerPoints, spellType); 

					if (powerPoints > maxPowerPoints) then
						powerfulSpell = spell;
						maxPowerPoints = powerPoints;
					end
				end
			end
		end
		return maxPowerPoints, powerfulSpell;

	else
		local powerfulSpell = HealPointsSpells:getSpell(HealPoints.db.char.power['spell']);
		local powerPoints;

		local spellType = powerfulSpell.class['type'];

		if (spellType == 'hot') then
			powerPoints = computeTotalHealingHot(powerfulSpell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellHaste, combatRegen);
		elseif (spellType == 'regrowth') then
			powerPoints = computeTotalHealingRegrowth(powerfulSpell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, spellHaste, combatRegen);
		elseif (spellType == 'lifebloom') then
			powerPoints = computeTotalHealingLifebloom(powerfulSpell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, spellHaste, combatRegen);
		else
			powerPoints = computeTotalHealingNormal(powerfulSpell, healing, 60 * HealPoints.db.char.power['duration'], mana, spellCrit, spellHaste, combatRegen);
		end

		HealPoints:LevelDebug(1, "         computePowerPoints: forced %s=%d (%s)", powerfulSpell.class['abbr'], powerPoints, spellType); 

		return powerPoints, powerfulSpell;
	end
end

--[[
	Caluclates one of the two terms in the HealPoints equation (HealPoints = PowerPoints+EndurancePoints
	Calculates how much healing can be done by an entire mana bar; takes mana regen into account
	@param maxMana: percent - The player's current mana (0..100), typically 0%
--]]
local function computeEndurancePoints(healing, spellCrit, spellHaste, maxMana, combatRegen)

	local mana = HealPoints.db.char.endurance['mana'] * maxMana / 100;

	--[[
	HealPoints:LevelDebug(1, "   computeEndurancePoints: healing=%d, spellCrit=%d, spellHaste=%d, maxMana=%d, combatRegen=%d, mana=%d", 
			healing, spellCrit, spellHaste, maxMana, combatRegen, mana);
	--]]

	if (HealPoints.db.char.endurance['auto'] == true) then
		local efficientSpell;
		local maxEndurancePoints = 0;
		for i = 1, table.getn(HealPointsSpells.SPELLTABLE), 1 do
			for j = 1, table.getn(HealPointsSpells.SPELLTABLE[i]), 1 do
				local spell = HealPointsSpells.SPELLTABLE[i][j];
				if (spell ~= nil and HealPointsSpells:getHighestSpellRank(spell.class.classID) == spell['rank'] and spell.class['type'] ~= 'instant') then
					local endurancePoints;
					local spellType = spell.class['type'];
					if (spell.class['type'] == 'hot') then
						endurancePoints = computeTotalHealingHot(spell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellHaste, combatRegen);
					elseif (spell.class['type'] == 'regrowth') then
						--HealPoints:LevelDebug(1, "      Spell %s, Type %s", spell.class['abbr'], "regrowth"); 
						endurancePoints = computeTotalHealingRegrowth(spell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, spellHaste, combatRegen);
					elseif (spell.class['type'] == 'lifebloom') then
						endurancePoints = computeTotalHealingLifebloom(spell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, spellHaste, combatRegen);
					else
						endurancePoints = computeTotalHealingNormal(spell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, spellHaste, combatRegen);
					end

					--HealPoints:LevelDebug(1, "      EndurancePoints: %s=%d (%s)", spell.class['abbr'], endurancePoints, spellType);

					if (endurancePoints > maxEndurancePoints) then
						efficientSpell = spell;
						maxEndurancePoints = endurancePoints;
					end
				end
			end
		end
		return maxEndurancePoints, efficientSpell;

	else
		local endurancePoints = 0;

		local efficientSpell = HealPointsSpells:getSpell(HealPoints.db.char.endurance['spell']);
		if (efficientSpell) then
			if (efficientSpell.class['type'] == 'hot') then
				endurancePoints = computeTotalHealingHot(efficientSpell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellHaste, combatRegen);
			elseif (efficientSpell.class['type'] == 'regrowth') then
				endurancePoints = computeTotalHealingRegrowth(efficientSpell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, spellHaste, combatRegen);
			elseif (efficientSpell.class['type'] == 'lifebloom') then
				endurancePoints = computeTotalHealingLifebloom(efficientSpell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, spellHaste, combatRegen);
			else
				endurancePoints = computeTotalHealingNormal(efficientSpell, healing, 60 * HealPoints.db.char.endurance['duration'], mana, spellCrit, spellHaste, combatRegen);
			end
		end;
		return endurancePoints, efficientSpell;
	end
end

--[[
	Calculates the player's
		healPoints, powerPoints, endurancePoints
	as well as the most powerfulSpell and the most efficient spell
	@param maxMana: number - The players current mana as a percent [0..100]. 
--]]
local function computeHealpoints(healing, spellCrit, spellHaste, maxMana, combatRegen)
	
	assert(healing, "computePowerPoints - healing is nil");
	assert(spellCrit, "computePowerPoints - spellCrit is nil");
	assert(spellHaste, "computePowerPoints - spellHaste is nil");
	assert(maxMana, "computePowerPoints - maxMana is nil");
	assert(combatRegen, "computePowerPoints - combatRegen is nil");
	--assert(manaRegenNormal, "computePowerPoints - manaRegenNormal is nil");

	--HealPoints:Print("computeHealPoints");

	--print(string.format("computeHealpoints.   combatRegen=%d", combatRegen));

	local powerPoints, powerfulSpell = computePowerPoints(healing, spellCrit, spellHaste, maxMana, combatRegen);
	local endurancePoints, efficientSpell = computeEndurancePoints(healing, spellCrit, spellHaste, maxMana, combatRegen);

	local healPoints = powerPoints + endurancePoints;

	return healPoints, powerPoints, endurancePoints, powerfulSpell, efficientSpell;
end

--[[
	Calculate the HealPoints gained per point of attributes
			intellect
			spirit
			spell power
			crit rating
			haste rating
	returns
		intellectValue
		spiritValue
		spellPowerValue
		critRatingValue
		hasteRatingValue
--]]
function HealPointsCalculator:ComputeHealPointsDelta(currentStats)
	assert(currentStats, "HealPointsCalculator:ComputeHealPointsDelta - currentStats cannot be nil")

	local iv, sv, spv, scrv, shrv;

	local newStats;

	--1 intellect
	local changes = { intellect = 1 };
	newStats = HealPointsCalculator:ApplyChanges(currentStats, changes);
	HealPointsCalculator:CalculateHealPoints(newStats);
	iv = (newStats.healPoints - currentStats.healPoints);

	--1 spirit
	local changes = { spirit = 1 };
	newStats = HealPointsCalculator:ApplyChanges(currentStats, changes);
	HealPointsCalculator:CalculateHealPoints(newStats);
	sv = (newStats.healPoints - currentStats.healPoints);

	--1 spell power
	local changes = { spellPower = 1 };
	newStats = HealPointsCalculator:ApplyChanges(currentStats, changes);
	HealPointsCalculator:CalculateHealPoints(newStats);
	spv = (newStats.healPoints - currentStats.healPoints);

	--1 spell crit rating
	local changes = { spellCritRating = 1 };
	newStats = HealPointsCalculator:ApplyChanges(currentStats, changes);
	HealPointsCalculator:CalculateHealPoints(newStats);
	scrv = (newStats.healPoints - currentStats.healPoints);

	--1 spell haste rating
	local changes = { spellHasteRating = 1 };
	newStats = HealPointsCalculator:ApplyChanges(currentStats, changes);
	HealPointsCalculator:CalculateHealPoints(newStats);
	shrv = (newStats.healPoints - currentStats.healPoints);

	return iv, sv, spv, scrv, shrv, currentStats.powerfulSpell, currentStats.efficientSpell;
end

--[[
	Get a table of the player's current (important) stats:
	
			data.playerLevel  		e.g. 85
			data.playerClass  		e.g. "Paladin"
			data.strength
			data.agility
			data.stamina
			data.intellect
			data.spirit
			data.mana				Player's maximum mana
			data.baseManaRegen 		Player's "base" mana regeneration (i.e. tooltip "Combat Regen")

			data.spellPower			Players tooltip "Spell Power" e.g. 3138
			data.spellCritChance	Player's tooltip "Crit Chance" e.g. 16.04%
			data.spellCritRating 	Player's spell "Crit Rating" e.g. Crit rating 263 (+4.36% crit chance) 
			data.spellHaste			Player's spell "Haste" (e.g. Haste +11.31%)
			data.spellHasteRating	Player's spell "Haste Rating" e.g. Haste rating 487 (+11.31% Haste)
			
			data.attackPower		Player's attack power, e.g. 399 (Increases damage with melee weapons by 28.5 damage per second.)
--]]
function HealPointsCalculator:GetCurrentPlayerStats(data)
	--assert(data)

	if (_cachedCurrentPlayerStats ~= nil) then
		HealPoints:LevelDebug(2, "HealPoints:GetCurrentPlayerStats() [Returning cached stats]");
		return _cachedCurrentPlayerStats;
	end;

	HealPoints:LevelDebug(1, "HealPoints:GetCurrentPlayerStats()");
	
	local result = data;
	if (result == nil) then
		result = {};
	end;

	local unit = "player"
	
	result.playerLevel = UnitLevel(unit)   --WoW API
--	result.playerHealth = UnitHealthMax(unit)   --WoW API
	result.playerClass = select(2, UnitClass(unit))	--   WoW API
	
	--Attributes
	--local stat, effectiveStat, posBuff, negBuff = UnitStat(unit, 1); --1=Strength
	result.strength =  UnitStat(unit, 1); --1=Strength   WoW API
	result.agility =   UnitStat(unit, 2); --2=Agility   WoW API
	result.stamina =   UnitStat(unit, 3); --3=Stamina   WoW API
	result.intellect = UnitStat(unit, 4); --4=Intellect   WoW API
	result.spirit =    UnitStat(unit, 5); --5=Spirit   WoW API
	
	-- Mana
	result.mana = UnitManaMax(unit); --WoW API
	result.baseMana = HealPointsUtil:GetBaseMana();

	local manaRegen, combatRegen = GetManaRegen(); --while not in combat, while in combat
	result.normalManaRegen = manaRegen;
	result.baseManaRegen = HealPointsUtil:GetBaseCombatManaRegen();
	result.combatManaRegen = combatRegen;
	result.combatManaRegenExtra = 0;
	result.manaCostModifier = 1.0;

	--Divine Plea option
	if (HealPoints.db.char.divineplea) then
		--Divine Plea: You gain 12% of your total mana over 9 seconds. 2 minute cooldown
		HealPoints:LevelDebug(3, "HealPointsCalculator:GetCurrentPlayerStats - Adding Divine Plea");
		if (result.playerClass == "PALADIN") then
			local manaRegenFromDivinePlea = result.mana*0.12/120;
			result.combatManaRegenExtra = result.combatManaRegenExtra + manaRegenFromDivinePlea;
		end;
	end;
	
	--Replenishment option
	if (HealPoints.db.char.replenishment) then
		--Replenishment: Regain 1% of maximum mana per 10 seconds over 15 seconds
		--15 ticks, each one being (mana*1% / 10)
		--or (mana * 1% / 10) per second
		HealPoints:LevelDebug(3, "HealPointsCalculator:GetCurrentPlayerStats - Adding Replenishment");
		local manaFromReplenishment = result.mana*0.01/10;
		result.combatManaRegenExtra = result.combatManaRegenExtra + manaFromReplenishment;
	end;

	result.combatManaRegen = result.combatManaRegen + result.combatManaRegenExtra;
	
	-- Spell
	result.spellPower = GetSpellBonusHealing(); --WoW API GetSpellBonusHealing (e.g. 3138)

	--print(string.format("GetCurrentPlayerStats: result.spellPower = %d", result.spellPower));

	result.spellCritRating = GetCombatRating(CR_CRIT_SPELL); --WoW API GetCombatRating(CR_CRIT_SPELL) = 263 ("Crit rating 263 (+4.36% crit chance)")
	--result.spellCritChanceBonus = GetCombatRatingBonus(CR_CRIT_SPELL); --WoW API GetCombatRatingBonus(CR_CRIT_SPELL) = 4.36 ("Crit rating 263 (+4.36% crit chance)")
	result.spellCritChance = GetSpellCritChance(2); --2=Holy  WoW API GetSpellCritChance(2) = 16.038 ("Crit Chance 16.04%")
	
	result.spellHasteRating = GetCombatRating(CR_HASTE_SPELL); --WoW API GetCombatRating(CR_HASTE_SPELL) = 487 ("Haste rating 487 (+11.31% haste)")
	local spellHasteBonus = GetCombatRatingBonus(CR_HASTE_SPELL);  --WoW API GetCombatRatingBonus(CR_HASTE_SPELL) = 11.3108 ("Haste rating 487 (+11.31% haste)")
	result.spellHaste = spellHasteBonus; --haste might come from other sources

	-- Melee
	result.attackPower = UnitAttackPower("unit"); --WoW API e.g. UnitAttackPower("player") = 399
	
	--[[
	-- Mana regen (casting, normal, %casting)
	local manaRegenNormal, manaRegenCasting, manaRegenNormalNew, manaRegenCastingNew, percentRegen = getCurrentAndNewManaRegen(0, 0, 0, 0);
	local manaRegenCastingDelta = manaRegenCastingNew - manaRegenCasting - _G["HealPointsVariables4InputBox"]:GetNumber();
	local manaRegenNormalDelta = manaRegenNormalNew - manaRegenNormal - _G["HealPointsVariables5InputBox"]:GetNumber();
	HealPointsCalculatorUI:setVariableStatInfo(4, manaRegenCasting, manaRegenCastingDelta, "%5.0f");
	HealPointsCalculatorUI:setVariableStatInfo(5, manaRegenNormal, manaRegenNormalDelta, "%5.0f");
	HealPointsCalculatorUI:setVariableStatInfo(6, 100 * percentRegen, 0, "%5.0f");
	--]]

	-- Healing spell power
	--local currentHealing, newHealing = getCurrentAndNewHealing(0, 0, 0, 0);
	--HealPointsCalculatorUI:setVariableStatInfo(7, currentHealing, newHealing - currentHealing - _G["HealPointsVariables7InputBox"):GetNumber(), "%5.0f"];

	-- Spell crit rating
	--HealPointsCalculatorUI:setVariableStatInfo(8, GetCombatRating(11), 0, "%5.0f");

	-- Spell haste rating
	--HealPointsCalculatorUI:setVariableStatInfo(9, GetCombatRating(20), 0, "%5.0f");
	
	_cachedCurrentPlayerStats = result;
	return result;
end;		


--[[
	Calculate result = playerStats+changes
		Changes can come from either
			a tooltip link -> HealPointsTooltips:GetChangesTable -> ApplyChanges
			the calculator -> HealPointsCalculatorUI:GetChanges -> ApplyChanges
		@param playerStats:table - a table of the player's stats to which the changes will be applied
		@param changes:table - a table of changes that will be applied to playerStats
	@returns: table - a player stats table containing the new player stats
--]]	
function HealPointsCalculator:ApplyChanges(playerStats, changes)
	assert(playerStats, "HealPointsCalculator:ApplyChanges - playerStats is nil");
	assert(changes, "HealPointsCalculator:ApplyChanges - changes is nil");

	--Get everything source has
	local results = Toolkit:CopyTable(playerStats);
	
	--[[
		data.playerLevel  		e.g. 85
		data.playerClass  		e.g. "Paladin"
		
		data.strength
		data.agility
		data.stamina
		data.intellect
		data.spirit
		
		data.mana				Player's maximum mana
		data.combatManaRegen 		Player's "base" mana regeneration (i.e. tooltip "Combat Regen")

		data.spellPower			Players tooltip "Spell Power" e.g. 3138
		data.spellCritRating 	Player's spell "Crit Rating" e.g. Crit rating 263 (+4.36% crit chance) 
		data.spellCritChance	Player's tooltip "Crit Chance" e.g. 16.04%
		data.spellHaste			Player's spell "Haste" (e.g. Haste +11.31%)
		data.spellHasteRating	Player's spell "Haste Rating" e.g. Haste rating 487 (+11.31% Haste)
			
		data.attackPower		Player's attack power, e.g. 399 (Increases damage with melee weapons by 28.5 damage per second.)
	--]]
	
	if (changes.strength ~= nil) then
		results.strength = playerStats.strength + (changes.strength or 0);
		changes.spellPower = (changes.spellPower or 0) + StatLogic:GetStatMod("ADD_HEAL_MOD_STR")*changes.strength;
		
		changes.strength = nil; --remove item
	end;
	
	changes.attackPower = (changes.attackPower or 0);
	if (changes.attackPower ~= nil) then
		results.attackPower = playerStats.attackPower + changes.attackPower;
		changes.spellPower = (changes.spellPower or 0) + StatLogic:GetStatMod("ADD_HEAL_MOD_AP")*changes.attackPower;

		changes.attackPower = nil; --remove item
	end;
	
	if (changes.agility ~= nil) then
		results.agility = playerStats.agility + changes.agility;
		changes.spellPower = (changes.spellPower or 0) + StatLogic:GetStatMod("ADD_HEAL_MOD_AGI")*changes.agility;

		changes.agility = nil; --remove item
	end;

	--results.stamina = playerStats.stamina + (changes.stamina or 0);

	--[[Combat Regen is a tricky thing. There are different ways of getting combat regen
		- Intellect -> Normal Regen -> adds to Combat Regen with meditation
		- Spirit -> Normal Regen -> adds to Combat Regen with meditation
		- 
	--]]
	local recalculateManaRegen = false;

	if (changes.intellect ~= nil) then
		results.intellect = playerStats.intellect + changes.intellect;

		changes.mana = (changes.mana or 0) + HealPointsUtil:GetManaFromInt(changes.intellect);
		changes.spellPower = (changes.spellPower or 0) + HealPointsUtil:GetSpellPowerFromInt(changes.intellect);
		changes.spellCritChance = (changes.spellCritChance or 0) + StatLogic:GetSpellCritFromInt(changes.intellect);

		recalculateManaRegen = true; --we changed our intellect, we have to recompute mana regen
		changes.intellect = nil; --remove item
	end;

	if (changes.spirit ~= nil) then
		results.spirit = playerStats.spirit + changes.spirit;

		recalculateManaRegen = true; --we changed our spirit, we have to recompute mana regen
		changes.spirit = nil; --remove item
	end;
	
	--local modMana = StatLogic:GetStatMod("MOD_MANA");
	--if (modMana ~= 0) then
	--end;
	
	if (changes.mana ~= nil) then
		results.mana = playerStats.mana + changes.mana;
		changes.spellPower = (changes.spellPower or 0) + StatLogic:GetStatMod("ADD_HEAL_MOD_MANA")*changes.mana;
		--changes.combatManaRegen = (changes.combatManaRegen or 0) + StatLogic:GetStatMod("ADD_COMBAT_MANA_REGEN_MOD_MANA")*changes.mana;
		--changes.combatManaRegen = (changes.combatManaRegen or 0)+ StatLogic:GetStatMod("ADD_COMBAT_MANA_REGEN_MOD_MANA_REGEN")*changes.mana;

		--recalculateManaRegen = true; --i don't think mana pool size affects regen; per se
		changes.mana = nil;
	end;

	HealPoints:LevelDebug(3, "results.mana: %s", results.mana);

	--MOD_MANA

	if (recalculateManaRegen) then
		--Intellect and Spirit can affect Mana Regen.
		--If the player has meditation then some portion of that normal regen can be added to Combat Regen.

		--If we're here then we're assuming that Spirit/Intellect are changing.
		--In order to calculate a new normalManaRegen we'll first subtract off the current regen from spi/int,
		--then add the new value from new spi/int

--		local baseManaRegen = HealPointsUtil:GetBaseCombatManaRegen(); --per second
		local manaRegenThroughSpiritBefore = StatLogic:GetNormalManaRegenFromSpi(playerStats.spirit, playerStats.intellect)/5;
		local manaRegenThroughSpiritAfter = StatLogic:GetNormalManaRegenFromSpi(results.spirit, results.intellect)/5;

		results.normalManaRegen = results.normalManaRegen - manaRegenThroughSpiritBefore + manaRegenThroughSpiritAfter;
		--changes.normalManaRegen = results.normalManaRegen - playerStats.normalManaRegen;

		--Now do a similar calculation for combat regen. But combat regen is only affected if the player has meditation
		local addCombatRegenModManaRegen = StatLogic:GetStatMod("ADD_COMBAT_MANA_REGEN_MOD_MANA_REGEN");
		if (addCombatRegenModManaRegen > 0) then
			manaRegenThroughSpiritBefore = manaRegenThroughSpiritBefore * addCombatRegenModManaRegen;	
			manaRegenThroughSpiritAfter = manaRegenThroughSpiritAfter * addCombatRegenModManaRegen;

			local newCombatManaRegen = results.combatManaRegen - manaRegenThroughSpiritBefore + manaRegenThroughSpiritAfter;

			--HealPoints:Print("Adjusting CombatRegen from %s to %s", results.combatManaRegen, newCombatManaRegen);
			results.combatManaRegen = newCombatManaRegen;
			--changes.combatManaRegen = (changes.combatManaRegen or 0) + (results.combatManaRegen - playerStats.normalManaRegen);
		end;	
--		HealPoints:LevelDebug(3, "manaRegenBefore: %s", manaRegenBefore);
		--HealPoints:LevelDebug(3, "manaRegenAfter: %s", manaRegenAfter);

		--changes.normalManaRegen = (manaRegenAfter - manaRegenBefore);

		--HealPoints:LevelDebug(3, "baseManaRegen: %s", baseManaRegen);


		--local combatRegenBefore = playerStats.combatManaRegen;
		--local combatRegenAfter = baseManaRegen + StatLogic:GetStatMod("ADD_COMBAT_MANA_REGEN_MOD_MANA_REGEN")*manaRegenThroughSpirit + playerStats.combatManaRegenExtra;
		--changes.combatManaRegen = (combatRegenAfter - combatRegenBefore);
	end;

	
	if (changes.combatManaRegen ~= nil) then
		--local combatRegenBefore = results.combatManaRegen;
		results.combatManaRegen = results.combatManaRegen + changes.combatManaRegen/5;
--		HealPoints:Print(string.format("Applying %s combat regen to existing regen of %s, for a total of %s",
					--changes.combatManaRegen, combatRegenBefore, results.combatManaRegen));
		changes.combatManaRegen = nil;
		HealPoints:Debug("results.combatManaRegin: "..results.combatManaRegen);
	end;

	if (changes.normalManaRegen ~= nil) then	
		results.normalManaRegen = playerStats.normalManaRegen + changes.normalManaRegen; 
		changes.normalManaRegen = nil;
		HealPoints:Debug("results.normalManaRegen: "..results.normalManaRegen);
	end;

	if (changes.combatManaRegenPercent ~= nil) then
--		results.combatManaRegenPercent = playerStats.combatManaRegenPercent + changes.combatManaRegenPercent;
		changes.combatManaRegenPercent = nil;
	end;

	if (changes.manaCost ~= nil) then
		results.manaCostModifier = playerStats.manaCostModifier + changes.manaCost;
		changes.manaCost = nil;
	end;
	
	HealPoints:LevelDebug(3, "changes.spellPower: %s", changes.spellPower);
	if (changes.spellPower ~= nil) then
		results.spellPower = playerStats.spellPower + changes.spellPower;
		changes.spellPower = nil;
	end;

	if (changes.spellCritRating ~= nil) then
		results.spellCritRating = playerStats.spellCritRating + changes.spellCritRating;
		changes.spellCritChance = (changes.spellCritChance or 0) + StatLogic:GetEffectFromRating(changes.spellCritRating, CR_CRIT_SPELL)/100.0;

		changes.spellCritRating = nil;
	end;

	if (changes.spellCritChance ~= nil) then
		results.spellCritChance = playerStats.spellCritChance + (changes.spellCritChance or 0);

		changes.spellCritChance = nil;
	end;

	if (changes.spellHasteRating ~= nil) then
		results.spellHasteRating = (playerStats.spellHasteRating or 0) + changes.spellHasteRating;

		changes.spellHasteRating = nil;

		results.spellHaste = StatLogic:GetEffectFromRating(results.spellHasteRating, CR_HASTE_SPELL);
	end;

	--Remove items from changes table that i know i can't use
	--results.MASTERY_RATING = nil;
	results.SPELL_DMG = nil;
	results.itemType = nil;
	results.link = nil;
	results.diffItemType1 = nil;
	results.diffItemType2 = nil;
	results.EXPERTISE_RATING = nil;
	results.ARMOR = nil;
	results.MELEE_CRIT_RATING = nil;
	results.MELEE_HASTE_RATING = nil;
	results.MELEE_HIT_RATING = nil;
	results.SPELL_HIT_RATING = nil;
	results.EMPTY_SOCKET_META = nil;
	results.EMPTY_SOCKET_RED = nil;
	results.EMPTY_SOCKET_YELLOW = nil;
	results.EMPTY_SOCKET_BLUE = nil;
	results.MAX_DAMAGE = nil;
	results.DPS = nil;
	results.DODGE_RATING = nil;
	results.PARRY_RATING = nil;
	results.ARMOR_BONUS = nil;
	results.SHADOW_RES = nil;
	results.FIRE_RES = nil;
	results.FROST_RES = nil;

	--Remove items from changes table that i know i can't use
	changes.MASTERY_RATING = nil;
	changes.SPELL_DMG = nil;
	changes.itemType = nil;
	changes.link = nil;
	changes.diffItemType1 = nil;
	changes.diffItemType2 = nil;
	changes.EXPERTISE_RATING = nil;
	changes.ARMOR = nil;
	changes.MELEE_CRIT_RATING = nil;
	changes.MELEE_HASTE_RATING = nil;
	changes.MELEE_HIT_RATING = nil;
	changes.SPELL_HIT_RATING = nil;
	changes.EMPTY_SOCKET_META = nil;
	changes.EMPTY_SOCKET_RED = nil;
	changes.EMPTY_SOCKET_YELLOW = nil;
	changes.EMPTY_SOCKET_BLUE = nil;
	changes.MAX_DAMAGE = nil;
	changes.DPS = nil;
	changes.DODGE_RATING = nil;
	changes.PARRY_RATING = nil;
	changes.ARMOR_BONUS = nil;
	changes.SHADOW_RES = nil;
	changes.FIRE_RES = nil;
	changes.FROST_RES = nil;


	HealPoints:LevelDebug(3, "changes.length %s", Toolkit:VarAsString(table.maxn(changes)));
	if next(changes) ~= nil then
		HealPoints:Debug("HealPointsCalculator:ApplyChanges - changes leftover: %s", Toolkit:VarAsString(changes));
	end;

	return results;
end;


--[[
	Based on the stats contained in passed playerStats, calculate the 
			HealPoints, PowerPoints and EndurancePoints, and other stats
		and write them to the same supplied playerStats object:
			- healPoints
			- powerPoints
			- endurancePoints
			- hpSec - (HP/second) Health Points healed per second by the selected spell
			- hpMana - (HP/mana) Health Points healed per point of mana
			- hpCast - (HP/cast) Health Points healed per cast of the selected spell
--]]
function HealPointsCalculator:CalculateHealPoints(playerStats, forceSpell)
	assert(playerStats, "HealPointsCalculator:CalculateHealPoints: Argument 'playerStats' is nil");
	--assert(forceSpell, "HealPointsCalculator:CalculateHealPoints - Argument 'spell' is nil")
		
	HealPoints:LevelDebug(3, "CalculateHealPoints(%s)", HealPointsSpells:GetSpellAbbr(forceSpell));
	
	--i need the spell in order to determine the points; the hp/sec, hp/cast, etc all come from the spell.
	--local manaRegenNormal = nil; --not ever needed  playerStats.normalManaRegen;
	local combatRegen = playerStats.combatManaRegen;

	local currentMana = playerStats.mana;
	local currentHealing = playerStats.spellPower;
	local currentSpellCrit = playerStats.spellCritChance;
	local currentSpellHaste = playerStats.spellHaste;

	--print(string.format("CalculateHealPoints: currentHealing=%d", currentHealing));

	assert(currentHealing, "CalculateHealPoints - currentHealing is nil");
	assert(currentSpellCrit, "CalculateHealPoints - currentSpellCrit is nil");
	assert(currentSpellHaste, "CalculateHealPoints - currentSpellHaste is nil");
	assert(currentMana, "CalculateHealPoints - currentMana is nil");
	assert(combatRegen, "CalculateHealPoints - combatRegen is nil");
	--assert(manaRegenNormal, "CalculateHealPoints - manaRegenNormal is nil");
	
	local healPoints, powerPoints, endurancePoints, powerfulSpell, efficientSpell =
			computeHealpoints(currentHealing, currentSpellCrit, currentSpellHaste, currentMana, combatRegen, forceSpell);

	playerStats.powerfulSpell = powerfulSpell;
	playerStats.efficientSpell = efficientSpell;
	
	self.HP_LastPowerfulCurrent = powerfulSpell;
--	self.HP_LastPowerfulNew = powerfulNew;
	self.HP_LastEfficientCurrent = efficientSpell;
--	self.HP_LastEfficientNew = efficientNew;

	HealPoints:LevelDebug(3, "CalculateHealPoints - powerfulSpell=%s", Toolkit:VarAsString(powerfulSpell));
	HealPoints:LevelDebug(3, "CalculateHealPoints - efficientSpell=%s", Toolkit:VarAsString(efficientSpell));

--	if (HealPoints_CalcFrame:IsVisible()) then
		--HealPointsCalculatorUI:setHealPoints(healedPowerful, healedPowerfulNew, healedEfficient, healedEfficientNew);
	--end
	--if (CharacterFrame:IsVisible()) then
--		HealPointsCharUI:setStats(healPointsCurrent, healedPowerful, healedEfficient, powerfulCurrent, efficientCurrent);
--	end
	
	if (forceSpell) then
		playerStats.hpSec = getHPSec(forceSpell, currentHealing, currentSpellCrit, currentSpellHaste);
		playerStats.hpMana = getHPMana(forceSpell, currentHealing, currentSpellCrit);
		playerStats.hpCast = getHPCast(forceSpell, currentHealing, currentSpellCrit);
		playerStats.hpManaBar = getHPTotalRegen(forceSpell, currentHealing, currentSpellCrit, currentSpellHaste, combatRegen, currentMana);
	end;

	playerStats.powerPoints = powerPoints;
	playerStats.endurancePoints = endurancePoints;
	playerStats.healPoints = healPoints;
end;

--[[
	Returns an object representing the players current stats and healpoint calculation results
	Note: This object might be calculated on the fly, or might be the cached result of a recent calculation
--]]
function HealPointsCalculator:GetCurrentHealPoints(forceSpell)

	local spellAbbr = HealPointsSpells:GetSpellAbbr(forceSpell);

	if (forceSpell == nil) then
		if (_cachedCurrentHealPoints) then
			HealPoints:LevelDebug(2, "GetCurrentHealPoints(%s) [returning cached values]", spellAbbr);
			return _cachedCurrentHealPoints;		
		end;
	end;

	HealPoints:LevelDebug(1, "GetCurrentHealPoints("..spellAbbr..")");

	local currentStats = HealPointsCalculator:GetCurrentPlayerStats();
	HealPointsCalculator:CalculateHealPoints(currentStats, forceSpell);

	if (forceSpell == nil) then
		_cachedCurrentHealPoints = currentStats;
	end;

	return currentStats;
end;

---Called by the controller HealPoints:ClearCache() when any cached stats are now invalid
function HealPointsCalculator:ClearCache(sender)
	HealPoints:LevelDebug(1, "HealPointsCalculator:ClearCache("..(sender or "")..")");
	_cachedCurrentHealPoints = nil;
	_cachedCurrentPlayerStats = nil;
end;